/// <summary>
/// Internal Clock Source Example for both Single and Multiple devices with the clock running at 1000 hz, with optionally using the AUX pin.
///
/// This example shows a general way of starting the SDK and then waiting for the Status events to indicate devices have been connected (StatusHandler).
/// Once devices are connected (BERTEC_DEVICES_READY), the StatusHandler function reads the channel names (bertec_GetDeviceChannelName) to determine
/// the index of the "FZ" channel name. This index value is used by the DataHandler function to display the running FZ value along with the device timestamp.
///
/// Both the StatusHandler and DataHandler functions are called from a secondary thread; the main WaitForStuff loop simply waits for the user to press a key
/// and then exits.
/// 
/// The StatusHandler code itself is important, since it shows how processing the DEVICES_READY signal is used to trigger the StartDataStream process.
/// In this example, the devices are set to INTERNAL CLOCKING mode where the first (or only) device is generating clock pulses and the others follow it.
///
/// While this example code is written to be used with only a single device (device index == 0), the code does handle multiple devices; it just ignores the others.
/// </summary>
using System;

namespace BertecExampleNET
{
	class InternalClockingFZReaderExample
	{
		BertecDeviceNET.BertecDevice theHandle = null;  // the library object
		int fzChannelIndex = -1;                        // set by StatusHandler and used by DataHandler and ImmediateDeviceDataHandler
		bool devicesAreaReady = false;                  // set after BERTEC_DEVICES_READY appears in the status and cleared whenever there is a problem; used by DataHandler
		bool demoImmediateDeviceDataHandler = false;    // for testing/demoing the single device data stream vs the combined data device stream.
		bool useAuxInsteadOfSync = false;					// set from the Run(bool) parameter: if TRUE, then the AUX line is used to control the start/stop of the sync flow

		static void Main(string[] args)
		{
			InternalClockingFZReaderExample example = new InternalClockingFZReaderExample();
			example.Run(false);
		}

		// This is an example "work" loop. The InitLibrary call starts the SDK, connects the event handlers, and starts things running in the background.
		// The middle part loops, waiting for a keypress (the "real work").
		// The final part tears everything down.
		// In your code, the InitLibrary might be part of the initial application setup, the keypress loop your main UI or whatever,
		// and the CloseLibrary is where you start shutting everything down (ex: your main application window or framework being destroyed, a finalizer, etc).
		public void Run(bool UseAux)
		{
			Console.WriteLine("Bertec example code #4. Press ESC or Space to exit.\n");

			useAuxInsteadOfSync = UseAux;

			if (InitLibrary() != 0)
				return;

			// Simply loop here waiting for the user to press escape or space

			int c = 0;
			while ((c = Console.ReadKey().KeyChar) != 3)
			{
				if (c == 27 || c == 32)
					break;
				System.Threading.Thread.Sleep(15);   // yield and appear to do something
			}

			CloseLibrary();
		}


		int InitLibrary()
		{
			devicesAreaReady = false;
			fzChannelIndex = -1;
			try
			{
				theHandle = new BertecDeviceNET.BertecDevice();
			}
			catch (System.Exception ex)
			{
				Console.WriteLine("Unable to initialize the Bertec Device Library (possible missing FTD2XX install).");
				return -1;
			}

			// This connects both the status change handler and the data handler to your functions.

			theHandle.OnStatus += StatusHandler;

			// While this example shows how to connect one event sink vs the other, you can absolutly connect both at the same time;
			// if you comment out the if-else you can see it happen on the console with the [I] flashing on and off and the text getting 'stomped on'.
			// This is caused by each event sink handler getting called at the same time by different thread processes, so they run concurrently.
			// Ex: you can get 2+ invokes at the same time on your OnImmediateDeviceData handler, all while your OnDataStream handler
			// is also concurrently processing data. If your event handlers must access shared resources (files, memory), you will need
			// to use mutex locking to marshal access - the example presented here doesn't, so that is why the Console output looks like it does.

			if (demoImmediateDeviceDataHandler)
				theHandle.OnImmediateDeviceData += ImmediateDeviceDataHandler;
			else
				theHandle.OnDataStream += DataHandler;

			// Start the device connection process in the background; StatusHandler and DataHandler will be called in separate threads as needed.
			theHandle.Start();

			return 0;
		}

		// Close the library and remove the event handlers. While .Dispose() is not 100% needed, it is recommended.
		void CloseLibrary()
		{
			devicesAreaReady = false;
			if (theHandle != null)
			{
				theHandle.OnStatus -= StatusHandler;
				theHandle.OnDataStream -= DataHandler;
				theHandle.OnImmediateDeviceData -= ImmediateDeviceDataHandler;
				theHandle.Stop();
				theHandle.Dispose();
			}
			theHandle = null;
		}


		// For long-term connections with multiple devices, it is suggested that periodically you tell the SDK to drop the connections and restart
		// to help with the multiple device synchronization and clock drift.
		// The timing if this will be application dependent - for example, this could be after every data collection "session" or on demand.
		// While the redetection is being preformed, you will not get any data via the DataHandler until the redetection is complete; any existing
		// data handler callbacks will remain and are not removed.
		void RedetectDevices()
		{
			theHandle.RedetectConnectedDevices();   // returns immediately
		}


		// The SDK supports the ability to upsample or downsample the data coming in from the devices. Typically, you will not want to do this and instead
		// provide some form of data manipulation in your own code.
		// Doing so will disable the SYNC and AUX pin modes, returning the device to SAMPLED mode; the SYNC and AUX data values will always be zero.
		// This feature works with any devices, even if they do not have the physical hardware for it.
		// Setting resampleRate to 0 will turn off resampling, and return the SDK to the hardware data rate of 1000hz.
		// Note that not all frequencies are available; the resulting value is equal to floor(8000/round(8000/input)).
		// More importantly, please note that data rate resampling can introduce delays in the data.
		void ChangeResamplingRate(int resampleRate)
		{
			int numDevices = theHandle.DeviceCount;
			for (int devIndex = 0; devIndex < numDevices; ++devIndex)
				theHandle.SetDataRateResampling(devIndex, resampleRate);
		}


		void FindFzIndex()
		{
			// A more robust process would look at each device's channel names and use separate FZ index values.
			int deviceCount = theHandle.DeviceCount;

			string[] channelNamesForDevice0 = theHandle.DeviceChannelNames(0);
			int channelCountForDevice0 = channelNamesForDevice0.Length;

			fzChannelIndex = -1; // assume it cannot be found
			for (int channelIndex = 0; channelIndex < channelCountForDevice0; ++channelIndex)
			{
				// If the channel name at this index is the FZ channel we're looking for, record off that index and exit.
				if (String.Equals(channelNamesForDevice0[channelIndex], "FZ", StringComparison.OrdinalIgnoreCase))
				{
					fzChannelIndex = channelIndex;
					break;
				}
			}
		}

		// The status handler will be called each time the _status value changes. The _status value will be one of the bertec_StatusErrors enums.
		// Not all enum values will be triggered through this; the ones that you can expect to see are implemented here. The rest are error values
		// returned from SDK api calls.
		void StatusHandler(BertecDeviceNET.StatusErrors status)
		{
			switch (status)
			{
				// This status value is emitted when the SDK starts probing the USB ports for connected devices.
				// At this point, there are no devices connected so the data values are reset. Your code should re-init
				// whatever processing function you're using, preparing it for use once devices are detected (BERTEC_DEVICES_READY)
				case BertecDeviceNET.StatusErrors.LOOKING_FOR_DEVICES:
					Console.WriteLine("\nSearching for connected devices");
					devicesAreaReady = false;
					fzChannelIndex = -1;
					break;

				// This status value is emitted when the SDK cannot find any connected USB devices.
				// While the data values were reset as before, we do it again here just to be sure.
				case BertecDeviceNET.StatusErrors.NO_DEVICES_FOUND:
					Console.WriteLine("\nNo devices found");
					devicesAreaReady = false;
					fzChannelIndex = -1;
					break;

				// This status value indicates that all connected devices have been successfully started and data is now being read in.
				// At this point, your DataHandler is already being called, but because the devicesAreaReady flag is set to FALSE,
				// the example DataHandler will not process the data yet. This gives the StatusHandler handler a chance to get the index
				// to the FZ channel for monitoring.
				case BertecDeviceNET.StatusErrors.DEVICES_READY:
					{
						Console.WriteLine("\nDevices found and ready");

						for (int devNum = 0; devNum < theHandle.DeviceCount; ++devNum)
						{
							Console.WriteLine("Plate serial {0}, {1}", theHandle.DeviceSerialNumber(devNum), theHandle.DeviceIDString(devNum));
						}

						// Starting with the 2.50 version of the SDK, the Library will no longer start delivering data the instant it detects devices;
						// this changes was done to improve how the system interacts with various end-user projects and to make it clear when data starts and stops.
						// The control block also allows various modes of operation - see the documentation on what all these are and how to use them.
						// For this example, the data stream is started in 'classical' non-synchronized mode.

						BertecDeviceNET.DataStreamControl streamControl = new BertecDeviceNET.DataStreamControl();
						streamControl.syncPinMode = BertecDeviceNET.DataStreamControl.SyncPinMode.INTCLOCK;  // internalClockSource will be the master clock
						streamControl.auxPinMode = BertecDeviceNET.DataStreamControl.AuxPinMode.NONE;      // and are passed in as part of the bertec_DataFrame structure.
						streamControl.deviceFilterBitmask = 0; // for non-zero values, any bit that is turned on will allow data through and off will not. This will affect the bertec_DataFrame structure.
						streamControl.internalClockSource = 0;
						streamControl.internalClockFrequency = 500; // the sync pin will be driven at this rate

						// If the UseAux parm was set, then instead of using the SYNC pin to determine the starting edge of the sync system, the SDK will instead
						// expect the AUX pin to be held in a LOW state and then triggered. Having all the device's AUX pins connected together and then into
						// an external control device that toggles (push button-like logic) will put the SDK into a start-stop mode (1st pulse: start data,
						// 2nd pulse: stop data. 3rd: start again, etc)
						if (useAuxInsteadOfSync)
						{
							streamControl.syncPinMode = BertecDeviceNET.DataStreamControl.SyncPinMode.INTAUXCONTROL;   // tells the SDK that the AUX pin will control the initial starting edge
							streamControl.auxPinMode = BertecDeviceNET.DataStreamControl.AuxPinMode.START_PULSE_ONOFF;  // tells the SDK that AUX pin pulses will toggle the data on/off
						}

						theHandle.StartDataStream(streamControl); // this will block until the device data stream is working; DataHandler will start being called when this exits successfully

						FindFzIndex();

						// Most of the time auto zeroing is desired, so turn that on once devices have been found.
						theHandle.AutoZeroing = true;

						devicesAreaReady = true;   // DataHandler will now process data
						break;
					}


				// These two errors indicate there is some sort of problem receiving data from the device.
				// The typical problem is that either the device was powered down or unplugged.
				// The SDK will restart the device detection routine, and you will get a BERTEC_LOOKING_FOR_DEVICES event shortly.
				// Your code should issue an update to the front end or log since, and handle data processing accordingly.

				// No data has been received in over 3 seconds, cable is probably unplugged
				case BertecDeviceNET.StatusErrors.NO_DATA_RECEIVED:
					Console.WriteLine("\nNo data being received");
					devicesAreaReady = false;
					fzChannelIndex = -1;
					break;

				// A communication error with a device has occurred - this usually occurs right after BERTEC_NO_DATA_RECEIVED.
				case BertecDeviceNET.StatusErrors.DEVICE_HAS_FAULTED:
					Console.WriteLine("\nDevice has faulted");
					devicesAreaReady = false;
					fzChannelIndex = -1;
					break;


				// These are advisory status values, and can be used to inform your user interface display.
				case BertecDeviceNET.StatusErrors.AUTOZEROSTATE_WORKING:
					Console.WriteLine("\nDetermining autozero");
					break;
				case BertecDeviceNET.StatusErrors.AUTOZEROSTATE_ZEROFOUND:
					Console.WriteLine("\nAutozero found");
					break;

				// For all the others, just show the status value.
				default:
					Console.WriteLine("\nStatus: {0}", status);
					break;
			}
		}

		// The DataHandler simply outputs the device's timestamp value along with the FZ value using the fzChannelIndex.
		// Each time this is called, the dataFrames collection will contain a single "row" of data for all the devices.
		// The dataFrames.Length value will be the same value as what calling BertecDeviceNET.DeviceCount would return.
		// For this example, only the first device is looked at; the code here loops through all the device data being
		// passed, but will only output for device index zero.
		void DataHandler(BertecDeviceNET.DataFrame[] dataFrames)
		{
			if (devicesAreaReady)
			{
				for (int deviceNumber = 0; deviceNumber < dataFrames.Length; ++deviceNumber)
				{
					// this is artificially contrived code; it's only done to show how to process other devices.
					if (deviceNumber == 0)
					{
						// make a reference alias to the device data block to keep things simple
						BertecDeviceNET.DataFrame deviceData = dataFrames[deviceNumber];

						// If the Unified Data Mode is turned OFF, you will get empty blocks of data for the device. Here we check for that.
						if (deviceData.forceData.Length > 0)
						{
							Console.Write("\r   {0} ", deviceData.timestamp);
							if (fzChannelIndex >= 0 && fzChannelIndex < deviceData.forceData.Length)
							{
								Console.Write(" {0}                  ", deviceData.forceData[fzChannelIndex]);
							}
							Console.Out.Flush(); ; // by design, .Write will only put out characters when newlines (\n) are printed.
						}
					}
				}
			}
		}

		// This event handler gets the force data from the plate after the calibration matrix and zero offsets have been applied, but
		// but before any additional calculations or processing (ex: filtering, resampling, computed channels). Each device will invoke
		// this handler when they get their data; you should process or buffer this as quick as possible and return. Device to device
		// ordering is not guaranteed, but data frame ordering within the same device is.
		// This handler is only turned on in this sample code file by setting the demoImmediateDeviceDataHandler, at the top of the file.
		void ImmediateDeviceDataHandler(int deviceIndex, string deviceUid, BertecDeviceNET.DataFrame deviceData)
		{
			if (devicesAreaReady)
			{
				Console.Write("[I]{0}, {1}, {2}", deviceIndex, deviceUid, deviceData.timestamp);
				if (fzChannelIndex >= 0 && fzChannelIndex < deviceData.forceData.Length)
				{
					Console.Write(": {0}                  ", deviceData.forceData[fzChannelIndex]);
				}
				Console.Out.Flush(); ; // by design, .Write will only put out characters when newlines (\n) are printed.
			}
		}
	}
}
